
#include <RotaryEncoder.h>
#include <hd44780.h>
#include <hd44780ioClass/hd44780_pinIO.h> // Arduino pin i/o class header

#define ROTARYSTEPS 1
#define ROTARYMIN 0
#define ROTARYMAX 250

//für die Mittelwertbildung
#define anzahlMittelWerte 10
float werte[anzahlMittelWerte];
int zaehlerMittelWerte=0;

//für die 2 Punkt Kalibrierung
/* first calibration point with real voltage (X1) and measured voltage (Y1) */
#define CAL_X1  (0.99f)
#define CAL_Y1  (0.96f)
/* second calibration point with real voltage (X2) and measured voltage (Y2) */
#define CAL_X2  (11.93f)
#define CAL_Y2  (11.96f)

/* factor for linear correction */
#define LINEAR_CORRECTION_M \
  (((CAL_Y2) - (CAL_Y1)) / ((CAL_X2) - (CAL_X1)))

/* offset for linear correction */
#define LINEAR_CORRECTION_N \
  (((CAL_Y1 * CAL_X2) - (CAL_Y2 * CAL_X1)) / (CAL_X2 - CAL_X1))

//Pin connected to clk shiftregister
int clockPin = 12;
////Pin connected to data shiftregister
int dataPin = 11;
////Pin connected to enable shiftregister
int enablePin = 10;

//Display IO
const int rs=8, en=9, db4=4, db5=5, db6=6, db7=7;      
hd44780_pinIO lcd(rs, en, db4, db5, db6, db7);

//Display LED enable
int dledPin = 3;

// LCD geometry
const int LCD_COLS = 16;
const int LCD_ROWS = 2;

// Setup a RoraryEncoder for pins A2 and A3:
RotaryEncoder encoder(A3, A2);

//Input button
const int button = 2;          // button on pin 2
volatile int button_stat = 0;
volatile int button_change = 0; 


void setup() {
  // initialisierung der seriellen Kommunikationsschnittstelle mit 9600 bits pro Sekunde
  Serial.begin(9600);
  while (!Serial) {};
  Serial.println("<Arduino is ready>");
  
  //init display
  pinMode(dledPin, OUTPUT);
  digitalWrite(dledPin, 1);
  lcd.begin(LCD_COLS, LCD_ROWS);
  lcd.print("CCCon RN20200325");
  delay(1000);
  lcd.clear();
  
  //set reference  
  analogReference(EXTERNAL);

  // You may have to modify the next 2 lines if using other pins than A2 and A3
  PCICR |= (1 << PCIE1);    // This enables Pin Change Interrupt 1 that covers the Analog input pins or Port C.
  PCMSK1 |= (1 << PCINT10) | (1 << PCINT11);  // This enables the interrupt for pin 2 and 3 of Port C.  

  pinMode(enablePin, OUTPUT);
  pinMode(dataPin, OUTPUT);
  pinMode(clockPin, OUTPUT);
  
  // set current to 0
  // sr out off
  digitalWrite(enablePin, 1);
  // data to sr
  shiftOut(dataPin, clockPin, 0);  
  // sr out on
  digitalWrite(enablePin, 0);
  // write to display OFF 
  lcd.setCursor(10, 0);
  lcd.print("OFF");


  pinMode(button, INPUT);       // define input for on/off button
  attachInterrupt(0, button_interrupt, CHANGE);   //define interrupt as CHANGE

  
}

// The Interrupt Service Routine for Pin Change Interrupt 1
// This routine will only be called on any signal change on A2 and A3: exactly where we need to check.
ISR(PCINT1_vect) {
  encoder.tick(); // just call tick() to check the state.
}

// Interrupt function for button
void button_interrupt() {           
  
  if (digitalRead(button) == 0){ 
    if (button_stat == 1){
      button_stat = 0;
    }
      
    else{
      button_stat = 1; 
    }
    button_change = 1;
  }
}

/* 20200325 erste Version constant current controller, funktional mit Ausgabe über RS232, der Strom wird nicht mehr über 8 Portpins ausgegeben sondern per Schieberegister
 so bleiben die Pins für das Display frei, geplant ist ein zwei Zeilen Textdisplay zur Anzeige der Spannung über die Last und den Sollstrom, sowie Aktivierung per Taste.
 20200326 Display eingebaut, Ausgabe funktioniert, on/off Button funktioniert -> kompliziert da jedes Schreiben in das Schieberegister ein flickern am Ausgang erzeugt,
 so wird nur im falle einer Änderung des Status auch in das Schieberegister geschrieben, deshalb die Variable "button_change".
 20200419 ADC auf externe Ref. umgestellt, Spannungsteiler auf 2,7M/220k geändert, Fehler in der Spannunsmessung <25mV von 0-16V, Zweipunktkalbrierung von mit Thomas 
 integriert, Fehler jetzt bei der Spannungsmessung kleiner 5mV 1Digit, kleinste zu messende Spannung auf 50mV festgelegt für eine stabile 0, ein Bit entspricht ca. 16mV.
 
*/

// Hauptschleife, welche endlos ausgeführt wird 
void loop() {
  static int pos = 0;
  static float temp = 0;

  int newPos = encoder.getPosition();
  //begrenze den Stellbereich auf 0-250
  if (newPos < ROTARYMIN) {
    encoder.setPosition(ROTARYMIN / ROTARYSTEPS);
    newPos = ROTARYMIN;
  } 
  else if (newPos > ROTARYMAX) {
    encoder.setPosition(ROTARYMAX / ROTARYSTEPS);
    newPos = ROTARYMAX;
  } 
  //lese Encoder ein
  if (pos != newPos) {
    Serial.print(newPos);
    Serial.println();
    pos = newPos;

    if (button_stat == 1){
      //current setting over shiftregister after the value has changed
      //sr out off
      digitalWrite(enablePin, 1);
      //data to sr
      shiftOut(dataPin, clockPin, newPos);  
      //sr out on
      digitalWrite(enablePin, 0);
    }
  }
  //current general on/off
  if (button_change == 1){
    button_change = 0; 
    //set current output on or off
    if (button_stat == 1){
      lcd.setCursor(10, 0);
      lcd.print("ON!");
  
      //current setting over shiftregister after the value has changed
      //sr out off
      digitalWrite(enablePin, 1);
      //data to sr
      shiftOut(dataPin, clockPin, newPos);  
      //sr out on
      digitalWrite(enablePin, 0);
    }
    
    else{
      lcd.setCursor(10, 0);
      lcd.print("OFF");
      
      //current setting 0, output off
      //sr out off
      digitalWrite(enablePin, 1);
      //data to sr
      shiftOut(dataPin, clockPin, 0);  
      //sr out on
      digitalWrite(enablePin, 0);  
    }
  }
  
  //clear the unused A from mA < 10 or < 100 
  if (pos < 100){
    lcd.setCursor(5, 0);
    lcd.print("  ");
  }
  if (pos < 10){
    lcd.setCursor(4, 0);
    lcd.print("  ");
  }

  // print the mA to display
  lcd.setCursor(0, 0);
  lcd.print(pos);
  lcd.print(" mA");
  
  // Lesen des Eingangs am analogen Pin A0:
  analogRead(A0); //first read dummy for better linearity
  delay(10);
  int sensorValue = analogRead(A0);
  
  // Umrechnung des Analogwertes (welcher von 0 bis 1023 reicht) in eine Spannung von (0V bis VRef):
  float voltage = sensorValue * ((1.23355 / 1023.0)*13.3);
  // Ausgabe des Wertes über die serielle Schnittstelle 

  //Zweipunktkalibrierung
  voltage = linar_correct(voltage);

  //Mittelwert bilden
  temp = mittelWert(voltage);   
  
  Serial.print(temp);
  Serial.println(" V");
  delay(100);
  
  //clear the unused V < 10 
  if (temp < 10){
    lcd.setCursor(6, 1);
    lcd.print("  ");
  }
  
  // print the Volage to display
  lcd.setCursor(0, 1);
  lcd.print(temp);
  lcd.print(" V");

  
  
}

void shiftOut(int myDataPin, int myClockPin, byte myDataOut) {
  // This shifts 8 bits out MSB first, 
  //on the rising edge of the clock,
  //clock idles low

  //internal function setup
  int i=0;
  int pinState;
  pinMode(myClockPin, OUTPUT);
  pinMode(myDataPin, OUTPUT);

  //clear everything out just in case to
  //prepare shift register for bit shifting
  digitalWrite(myDataPin, 0);
  digitalWrite(myClockPin, 0);

  //for each bit in the byte myDataOut
  //NOTICE THAT WE ARE COUNTING DOWN in our for loop
  //This means that %00000001 or "1" will go through such
  //that it will be pin Q0 that lights. 
  for (i=7; i>=0; i--)  {
    digitalWrite(myClockPin, 0);

    //if the value passed to myDataOut and a bitmask result 
    // true then... so if we are at i=6 and our value is
    // %11010100 it would the code compares it to %01000000 
    // and proceeds to set pinState to 1.
    if ( myDataOut & (1<<i) ) {
      pinState= 1;
    }
    else {  
      pinState= 0;
    }

    //Sets the pin to HIGH or LOW depending on pinState
    digitalWrite(myDataPin, pinState);
    //register shifts bits on upstroke of clock pin  
    digitalWrite(myClockPin, 1);
    //zero the data pin after shift to prevent bleed through
    digitalWrite(myDataPin, 0);
  }

  //stop shifting
  digitalWrite(myClockPin, 0);
}

//Zweipunktkalibrierung mit stabiler 0
float linar_correct(float val)
{
  if (val < 0.05){
    return 0;
  }
  else{
    return (val - LINEAR_CORRECTION_N) / LINEAR_CORRECTION_M;
  }
  
}


float mittelWert(float neuerWert) {
// neuen float Wert eingeben und den Mittelwert als float zurück geben
//
// Matthias Busse 9.2.2016 Version 1.0
float mittel, summe=0;

  werte[zaehlerMittelWerte] = neuerWert;
  for(int k=0; k < anzahlMittelWerte; k++) summe += werte[k];
  mittel=(float) summe / anzahlMittelWerte; 
  zaehlerMittelWerte++;
  if(zaehlerMittelWerte >= anzahlMittelWerte) zaehlerMittelWerte=0;
  return mittel;
}
